


/****************************************************************************
;
; NAME: osata.c
;
;
; DESCRIPTION:
;
;       Linux interface ATA functions
;
;
; REVISION HISTORY:
;
;       $Log: /ESASHBA2/Linux/driver/osata.c $
; 
; 1     10/25/10 4:42p Mseier
; Initial entry.
; 
; 
; 
;
; THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
; ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
; WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
; PURSUANT TO WHICH IT WAS FURNISHED.
;
; COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2010
; ALL RIGHTS RESERVED.
;
;***************************************************************************/

#ifdef __VMKLNX__
#include "oscommon.h"
#else
#include "os.h"
#endif
#include "oswrap.h"
#include "osata.h"


/****************************************************************************

    FUNCTION:   int OsAtaSendCommand( void* req_parm, 
                                      unsigned char* cmd, 
                                      unsigned char* buffer,
                                      int size    )

    PURPOSE:    Sends a SCSI ATA_PASSTHROUGH command via the Linux kernel.

    ARGUMENTS:  req_parm- Either a scsi_request or Scsi_Device, depending on
                          the kernel version.
                cmd     - ATA_PASSTHROUGH command
                buffer  - Data buffer attached to the request
                size    - Length of the data buffer

    RETURNS:    0 if the command is sent successfully, else returns -EIO.

****************************************************************************/

int OsAtaSendCommand( void* req_parm, 
                      unsigned char* cmd,
                      unsigned char* buffer,
                      int size )
{
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,11)
    struct scsi_request* scsi_req = (struct scsi_request*) req_parm;
    scsi_req->sr_data_direction = (size) ? DMA_FROM_DEVICE : DMA_NONE;

    scsi_wait_req (scsi_req, cmd,
                   buffer, size,
                   20 * HZ, 5);

    if (scsi_req->sr_result)
    {
        DEBUG_ALL (("scsi_wait_req failed (%X)\n", scsi_req->sr_result));
        return( -EIO );
    }
#else
    Scsi_Device* scsi_req = (Scsi_Device*) req_parm;
    if (scsi_execute_req ( scsi_req, cmd,
                          (size) ? DMA_FROM_DEVICE : DMA_NONE,
                          buffer, size,
                          NULL, 20 * HZ, 5
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,29)
                          , NULL
#endif
        ))
    {
        DEBUG_ALL (("scsi_execute_req failed\n"));
        return( -EIO );
    }
#endif

    return( 0 );
}


/****************************************************************************

    FUNCTION:   int OsAtaSetupDrvCmd( unsigned char* cmd_buf, 
                                      unsigned char* ata_passthru_args,
                                      void* req_parm,
                                      void* args )

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_DRIVE_CMD
                Ioctl.

    ARGUMENTS:  cmd_buf - ATA_PASSTHROUGH command to fill in
                ata_passthru_args - Arguments to the Ioctl
                req_parm- Either a scsi_request or Scsi_Device, depending on
                          the kernel version.
                args    - Pointer to the user-space ATA Ioctl arguments buffer

    RETURNS:    OSIOCTL_ALL_DONE if the command is sent successfully, else
                returns an errno value.

****************************************************************************/

int OsAtaSetupDrvCmd( unsigned char* cmd_buf, 
                      unsigned char* ata_passthru_args, 
                      void* req_parm,
                      void* args )
{
    int rez;
    int buf_size = ata_passthru_args [3] * 512;
    unsigned char* ata_pass_thru_buffer = NULL;

    /* Set up the command structure */

    cmd_buf [0] = 0xa1; /* ATA PASS-THROUGH 12 */
                                  
    /* ATA Register:  */

    cmd_buf [9] = ata_passthru_args [0]; /* Command Code   */
    cmd_buf [5] = ata_passthru_args [1]; /* Sector # (LBA) */
    cmd_buf [3] = ata_passthru_args [2]; /* Features       */
    cmd_buf [4] = ata_passthru_args [3]; /* Sector count   */

    if (ata_passthru_args [0] == 0xb0)
    {
        /* This is a SMART command, put     */
        /* magic numbers in LBA Med, LBA Hi */

        cmd_buf [6] = 0x4f;
        cmd_buf [7] = 0xc2;
    }

    /* Allocate a buffer */

    if (buf_size)
    {
        ata_pass_thru_buffer = kmalloc (buf_size,
                                        GFP_KERNEL);

        if (ata_pass_thru_buffer == NULL)
        {
            rez = -ENOMEM;
            DEBUG_ALL (("kmalloc (ata_pass_thru_buffer) failed\n"));
AtaPassThruError:
            if (ata_pass_thru_buffer) kfree (ata_pass_thru_buffer);
            return rez;
        }

        cmd_buf [1] = 0x08;  /* Protocol - PIO Data In */
        cmd_buf [2] = 0x0E;
    }
    else
        cmd_buf [1] = 0x06;  /* Protocol - Non-data */ 

    rez = OsAtaSendCommand( req_parm, 
                            cmd_buf, 
                            ata_pass_thru_buffer, 
                            buf_size );

    if ( rez < 0 )
        goto AtaPassThruError;
    else if ( buf_size )
    {
        /* Save the response data back to the user's buffer. */

        if (__copy_to_user ((void *) (args + 4),
                            ata_pass_thru_buffer,
                            buf_size) != 0)
        {
            DEBUG_ALL (("ioctl_handler copy_to_user didn't copy it all\n"));
            rez = -EFAULT;
            goto AtaPassThruError;
        }

        kfree (ata_pass_thru_buffer);
    }

    return ( rez );
}


/****************************************************************************

    FUNCTION:   int OsAtaSetupDrvTask( unsigned char* cmd_buf, 
                                       unsigned char* ata_passthru_args,
                                       void* req_parm )

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_DRIVE_TASK
                Ioctl.

    ARGUMENTS:  cmd_buf - ATA_PASSTHROUGH command to fill in
                ata_passthru_args - Arguments to the Ioctl
                req_parm- Either a scsi_request or Scsi_Device, depending on
                          the kernel version.

    RETURNS:    OSIOCTL_ALL_DONE if the command is sent successfully, else
                returns an errno value.

****************************************************************************/

int OsAtaSetupDrvTask( unsigned char* cmd_buf, 
                       unsigned char* ata_passthru_args, 
                       void* req_parm )
{
    int rez;

    /* ATA Register:  */

    cmd_buf [0] = 0xa1; /* ATA PASS-THROUGH 12 */
    cmd_buf [9] = ata_passthru_args [0]; /* Command Code   */
    cmd_buf [3] = ata_passthru_args [1]; /* Features       */
    cmd_buf [4] = ata_passthru_args [2]; /* Sector count   */
    cmd_buf [5] = ata_passthru_args [3]; /* Sector # (LBA) */
    cmd_buf [6] = ata_passthru_args [4]; /* Cyl low        */
    cmd_buf [7] = ata_passthru_args [5]; /* Cyl high       */
    cmd_buf [8] = ata_passthru_args [6]; /* Device head    */
    cmd_buf [1] = 0x06;     /* Protocol - Non-data */

    rez = OsAtaSendCommand( req_parm, 
                            cmd_buf, 
                            NULL, 
                            0 );

    return ( rez );
}


/****************************************************************************

    FUNCTION:   int OsAtaSetupSetPIOMode(   unsigned char* cmd_buf, 
                                            long val, 
                                            void* req_parm )

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_SET_PIO_MODE
                Ioctl.

    ARGUMENTS:  cmd_buf - ATA_PASSTHROUGH command to fill in
                val - Value the PIO mode is to be set to.
                req_parm- Either a scsi_request or Scsi_Device, depending on
                          the kernel version.

    RETURNS:    OSIOCTL_ALL_DONE if the command is sent successfully, else
                returns an errno value.

****************************************************************************/

int OsAtaSetupSetPIOMode(   unsigned char* cmd_buf, 
                            long val, 
                            void* req_parm )
{
    int rez;

    if ( val > 255
      || val < 0 )
        return ( -EINVAL );

    /* ATA Register:  */

    cmd_buf [0] = 0xa1; /* ATA PASS-THROUGH 12                      */
    cmd_buf [9] = 0xef; /* Command Code- SET FEATURES               */
    cmd_buf [3] = 3;    /* Features- Set transfer mode              */
    cmd_buf [4] = (unsigned char) val;  /* Sector count- PIO mode   */
    cmd_buf [1] = 0x06;     /* Protocol - Non-data */

    rez = OsAtaSendCommand( req_parm, 
                            cmd_buf, 
                            NULL, 
                            0 );

    return ( rez );
}


/****************************************************************************

    FUNCTION:   int OsAtaSetupSetMultcount( unsigned char* cmd_buf, 
                                            int val, 
                                            void* req_parm )

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_SET_MULTCOUNT
                Ioctl.

    ARGUMENTS:  cmd_buf - ATA_PASSTHROUGH command to fill in
                val - Value the multcount is to be set to.
                req_parm- Either a scsi_request or Scsi_Device, depending on
                          the kernel version.

    RETURNS:    OSIOCTL_ALL_DONE if the command is sent successfully, else
                returns an errno value.

****************************************************************************/

int OsAtaSetupSetMultcount( unsigned char* cmd_buf, 
                            int val, 
                            void* req_parm )
{
    int rez;

    if ( val > 255
      || val < 0 )
        return ( -EINVAL );

    /* ATA Register:  */

    cmd_buf [0] = 0xa1; /* ATA PASS-THROUGH 12              */
    cmd_buf [9] = 0xc6; /* Command Code- SET MULTIPLE MODE  */
    cmd_buf [3] = 3;    /* Features- Set transfer mode      */
    cmd_buf [4] = (unsigned char) val;  /* Count- Block count in this case  */
    cmd_buf [1] = 0x06;     /* Protocol - Non-data */

    rez = OsAtaSendCommand( req_parm, 
                            cmd_buf, 
                            NULL, 
                            0 );

    return ( rez );
}


/****************************************************************************

    FUNCTION:   int OsAtaSetupDrvReset( unsigned char* ata_pass_thru_cmd,
                                        void* req_parm )

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_DRIVE_RESET
                Ioctl.

    ARGUMENTS:  cmd_buf - ATA_PASSTHROUGH command to fill in
                req_parm- Either a scsi_request or Scsi_Device, depending on
                          the kernel version.

    RETURNS:    OSIOCTL_ALL_DONE if the command is sent successfully, else
                returns an errno value.

****************************************************************************/

int OsAtaSetupDrvReset( unsigned char* cmd_buf,
                        void* req_parm )
{
    int rez;
    
    /* ATA Register:  */
    /* Fill in a DEVICE RESET command */

    cmd_buf [0] = 0xA1; /* ATA PASS-THROUGH 12 */
    cmd_buf [1] = 0x02; /* Protocol - Soft Reset */

    rez = OsAtaSendCommand( req_parm, 
                            cmd_buf, 
                            NULL, 
                            0 );

    return ( rez );
}


/****************************************************************************

    FUNCTION:   int OsAtaSetupGetMultcount( unsigned char* cmd_buf,
                                            void* req_parm,
                                            void* args )

    PURPOSE:    Fills in an ATA_PASSTHROUGH command for an HDIO_GET_MULTICOUNT
                Ioctl.

    ARGUMENTS:  cmd_buf - ATA_PASSTHROUGH command to fill in
                req_parm- Either a scsi_request or Scsi_Device, depending on
                          the kernel version.
                args    - Pointer to the user-space ATA Ioctl arguments buffer

    RETURNS:    OSIOCTL_ALL_DONE if the command is sent successfully, else
                returns an errno value.

****************************************************************************/

int OsAtaSetupGetMultcount( unsigned char* cmd_buf, 
                            void* req_parm,
                            void* args )
{
    int  rez;
    int  buf_size   = 512;
    long multcount;
    u16* ata_pass_thru_buffer = kmalloc (buf_size, GFP_KERNEL);

    if (ata_pass_thru_buffer == NULL)
    {
        rez = -ENOMEM;
        DEBUG_ALL (("kmalloc (ata_pass_thru_buffer) failed\n"));
AtaPassThruError:
        if (ata_pass_thru_buffer) kfree (ata_pass_thru_buffer);
        return rez;
    }

    /* All we do here is send out an IDENTIFY DEVICE command, and return    */
    /* the byte that specifies the MultCount.                               */

    /* Set up the command structure */

    cmd_buf [0] = 0xa1; /* ATA PASS-THROUGH 12 */
                                  
    /* ATA Register:  */

    cmd_buf [9] = 0xEC; /* Command Code- Identify Device */
    cmd_buf [1] = 0x08; /* Protocol - PIO Data In */
    cmd_buf [2] = 0x0E; /* Flags- we are sending blocks from drive to adap */
    cmd_buf [4] = 1;    /* Sector Count- Block count expected in this case */

    rez = OsAtaSendCommand( req_parm,
                            cmd_buf,
                            (unsigned char*) ata_pass_thru_buffer,
                            buf_size );

    if ( rez < 0 )
        goto AtaPassThruError;
    else
    {
        /* Copy the number of logical sectors transferred on a READ/WRITE   */
        /* MULTIPLE command back to the user's buffer.                      */

        /* First check that the count is valid */

        if ( (ata_pass_thru_buffer[59] & 0xff00) == 0x100 )
            multcount = ata_pass_thru_buffer[59] & 0xff;
        else
            multcount = 0;

        DEBUG_ALL (( "field: %x  multcount: %ld\n", ata_pass_thru_buffer[59],
                     multcount ));

        /* Then copy the value to user-space */

        if (__copy_to_user ( args,
                             &multcount,
                             sizeof(long) ) != 0)
        {
            DEBUG_ALL (("ioctl_handler copy_to_user didn't copy it all\n"));
            rez = -EFAULT;
            goto AtaPassThruError;
        }

        kfree (ata_pass_thru_buffer);

        return ( rez );
    }
}


/****************************************************************************

    FUNCTION:   int OsAtaGetIoctlSize (int cmd)

    PURPOSE:    Returns size of data argument for specified ATA IOCTL

    ARGUMENTS:  cmd - ATA IOCTL command code

    RETURNS:    Size of data argument for specified IOCTL if we support it,
                else return -1.

****************************************************************************/

int OsAtaGetIoctlSize (int cmd)
{
    /* Some processes (eg. hdparm) may send us ioctls that are only meant   */
    /* for IDE devices, or are so old that they are not supported by        */
    /* libata.  Our policy is to support SG_IO, all the HDIO ioctls that    */
    /* libata supports, and all the ioctls that can be sent through an ATA  */
    /* PASSTHROUGH command. We mark the rest as not implemented so they     */
    /* will be rejected without further processing.                         */

    int size;

    switch (cmd)
    {
        case HDIO_GET_MULTCOUNT:
            size = sizeof(void*);
            break;
        case HDIO_SET_MULTCOUNT:
            size = sizeof(int);
            break;
        case HDIO_SET_PIO_MODE:
            size = sizeof(long);
            break;
        case HDIO_DRIVE_RESET:
            size = 0;
            break;
        case HDIO_DRIVE_CMD:
            size = 4;
            break;
        case HDIO_DRIVE_TASK:
            size = 7;
            break;
        case HDIO_DRIVE_TASKFILE:
        case HDIO_GET_IDENTITY:
        case HDIO_SET_32BIT:
        case HDIO_GET_32BIT:
        case HDIO_GET_BUSSTATE:
        case HDIO_SET_BUSSTATE:
        case HDIO_GET_NOWERR:
        case HDIO_SET_NOWERR:
        case HDIO_GET_WCACHE:
        case HDIO_SET_WCACHE:
        case HDIO_GET_DMA:
        case HDIO_SET_DMA:
        case HDIO_GET_KEEPSETTINGS:
        case HDIO_SET_KEEPSETTINGS:
        case HDIO_GET_UNMASKINTR:
        case HDIO_SET_UNMASKINTR:
            size = ATTO_NOT_IMPLMNTD;
            break;
        default:
            size = -1;
            break;
    }

    return size;
}


/****************************************************************************

    FUNCTION:   int OsAtaProcessIoctl( int cmd, void* ioctl_data, 
                                       void* hostdata, void* arg )

    PURPOSE:    Convert an ATA Ioctl from the Linux interface into a an ATA
                passthrough command, send it out to the drive, and return the
                results to the Linux user layer.

    ARGUMENTS:  cmd - ATA IOCTL command code
                ioctl_data - Pointer to kernel space Ioctl data buffer
                sd  - The Scsi_Device host device
                arg - Pointer to the user-space ATA Ioctl arguments buffer
                
    RETURNS:    OSIOCTL_ALL_DONE if IOCTL processed and completed successfully, 
                OSIOCTL_COPY_BACK if OSIOCTL_ALL_DONE and ioctl_data needs to 
                be copied back into the args.
                If unsuccessful, return the approprate errno.

****************************************************************************/

int OsAtaProcessIoctl( int cmd, void* ioctl_data, Scsi_Device* sd, void* arg )
{
    int rez = OSIOCTL_ALL_DONE;
    unsigned char  ata_pass_thru_cmd [MAX_COMMAND_SIZE];
    unsigned char *ata_pass_thru_args = (unsigned char *) ioctl_data;
    void* req_parm;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,11)
    struct scsi_request *scsi_req;

    if (!(scsi_req = scsi_allocate_request (sd, GFP_KERNEL)))
    {
        DEBUG_ALL (("scsi_allocate_request failed\n"));
        vfree (ioctl_data);
        return -EINTR;
    }

    req_parm = (void*) scsi_req;
#else
    req_parm = (void*) sd;
#endif

    memset (ata_pass_thru_cmd, 0, sizeof (ata_pass_thru_cmd));

    switch (cmd)
    {
        case HDIO_GET_MULTCOUNT:
            rez = OsAtaSetupGetMultcount( ata_pass_thru_cmd,
                                          req_parm,
                                          arg );
            break;
        case HDIO_SET_MULTCOUNT:
            rez = OsAtaSetupSetMultcount( ata_pass_thru_cmd,
                                          (int) *ata_pass_thru_args,
                                          req_parm );
            break;
        case HDIO_SET_PIO_MODE:
            rez = OsAtaSetupSetPIOMode( ata_pass_thru_cmd,
                                        (long) *ata_pass_thru_args,
                                        req_parm );
                                        
            break;
        case HDIO_DRIVE_RESET:
            rez = OsAtaSetupDrvReset( ata_pass_thru_cmd,
                                      req_parm );                                      
            break;
        case HDIO_DRIVE_CMD:
            rez = OsAtaSetupDrvCmd( ata_pass_thru_cmd, 
                                    ata_pass_thru_args,
                                    req_parm,
                                    arg );
            break;
        case HDIO_DRIVE_TASK: /* Also covers HDIO_DRIVE_CMD_AEB */
            rez = OsAtaSetupDrvTask( ata_pass_thru_cmd,
                                     ata_pass_thru_args,
                                     req_parm );
            break;
        default:
            /* We're getting an ATA IOCTL that we haven't implemented */
            rez = ATTO_NOT_IMPLMNTD;
            break;
    }

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,11)
    scsi_release_request (scsi_req);
#endif

    return( rez );
}
