


/****************************************************************************
;
; NAME: ostransport.c
;
;
; DESCRIPTION:
;
;       Linux driver SAS transport interface
;
;
; REVISION HISTORY:
;
;       $Log: /EsasHba2/Linux/driver/ostransport.c $
; 
; 21    5/30/13 9:05a Pcullen
; Another modification of the device listing in our proc file for Hitachi
; (IR 8703)
; 
; 20    1/17/13 3:23p Pcullen
; IR 8307: Change order of devices in driver proc file.
; 
; 19    1/09/12 4:24p Pcullen
; IR7845	Kernel Panic in CentOS 5.3
; IR7864	Need to be able to report enclosure IDs and slot information in
; SAS transport layer.
; NOTE: ostransport.[ch] were branched to use the new discovery code in
; the HW layer. That code is only in the ESASHBA2 HW layer at this point,
; and since ostransport is common between ESASHBA2 and ESASHBA3, it had
; to be branched. The branch should be merged once the ESASHBA3 HW layer
; contains the same discovery functionality.
; 
; 17    1/04/12 2:03p Jseba
; Fix support for ESAS3HBA
; 
; 16    11/14/11 11:33a Pcullen
; Added support for get_bay_identifier is SAS Transport for ESAS2HBA
; only.
; 
; 15    11/10/11 10:46a Jseba
; Add PHY discovery retries; fix a few potential NULL pointer
; dereferences and resource leaks 
; 
; 14    10/26/11 3:16p Pcullen
; Instrumented SAS Transport layer with ATTO_LOG; Fix for IR 7845: kernel
; panic after drive pull/reinsert.
; 
; 13    9/16/11 3:00p Jseba
; Fix a panic when a non-expander, non-host device adds a device
; 
; 12    4/20/11 1:15p Mseier
; Added check for virtual device along with connected phy check
; 
; 11    11/24/10 1:29p Pcullen
; Fix for IR7366 - initialize topology host_no to -1 and search all
; topologies when looking for one by host number.
; 
; 10    9/17/10 11:23a Mseier
; IR 7270: Changed #ifdefs so Linux bus ID is not referenced in kernel
; versions 2.6.30 and up, previously 2.6.31 and up.
; 
; 9     6/18/10 10:06a Pcullen
; Fix for IR7129: kernel panic with 6Gb - Moved list head initialization
; further up in the call stack so it always happens when a device is made
; into an expander.
; 
; 8     6/16/10 8:55a Pcullen
; Fixed list initialization of port_list so that it occurs event in the
; SMP Report General command fails, or the mem alloc fails (less likely).
; 
; 7     5/27/10 4:20p Pcullen
; Esas2hba port to VMware 4.x; ifdef out the unimplemented
; sas_port_free() function calls
; 
; 6     5/19/10 2:18p Pcullen
; Change one last false to 0
; 
; 5     5/19/10 1:23p Pcullen
; Remove usage of bool, true and false.
; 
; 4     4/22/10 8:16a Pcullen
; Implementation of Linux SAS Transport
; 
; 3     11/24/09 4:21p Pcullen
; Add get_target_sas_identifier entry point to satisfy VMkernel version
; of SAS transport.
; 
; 2     11/24/09 11:05a Pcullen
; Initial checkin from VMware 4.0 porting
; 
; 1     11/20/09 4:19p Pcullen
; Rudimentary Linux SAS Transport implementation
; 
;
;
; THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
; ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
; WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
; PURSUANT TO WHICH IT WAS FURNISHED.
;
; COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2003-2009
; ALL RIGHTS RESERVED.
;
;***************************************************************************/


#ifdef __VMKLNX__
#include "oscommon.h"
#else
#include "os.h"
#endif

#include "ostransport.h"

#ifdef ATTO_TRANSPORT

#ifdef ATTO_SAS_PORTS

/* Each adapter gets it's own topology */
ESASTOPOLOGY esas_topology[32];
#endif /* ATTO_SAS_PORTS */


#ifndef list_first_entry
#define list_first_entry(ptr, type, member) list_entry((ptr)->next, type, member)
#endif

#ifdef __VMKLNX__
static int esashba_get_target_sas_identifier (struct scsi_target *starget, u64 *sas_id);
#endif

/* Define Transport methods based on kernel availability */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
int esashba_get_linkerrors (struct sas_phy *phy)
{

    return -ENXIO;
}


int esashba_phy_reset (struct sas_phy *phy, int param)
{

    return -ENXIO;
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
int esashba_get_bay_identifier (struct sas_rphy *rphy)
{
    int bay = 0;

#if defined (ESAS2HBA) || defined (ESAS3HBA)
    struct Scsi_Host *shost;
    PESASTOPOLOGY etop;
    PESASDEVICE edev;

    ATTO_LOG (KERN_DEBUG, ATTO_LOG_TOPOL,
              "esashba_get_bay_identifier(rphy:0x%llx) called\n",
              rphy->identify.sas_address);

    shost = rphy_to_shost(rphy);
    if (!shost)
    {
        ATTO_LOG (KERN_WARN, ATTO_LOG_TOPOL,
                  "esashba_get_bay_identifier: invalid scsi host for rphy 0x%llx\n",
                  rphy->identify.sas_address);
        return -ENXIO;
    }
    ATTO_LOG (KERN_DEBUG, ATTO_LOG_TOPOL,
              "esashba_get_bay_identifier: found host%d for rphy\n",
              shost->host_no);

    etop = esas_topology_by_host_no (shost->host_no);
    if (!etop)
    {
        ATTO_LOG (KERN_WARN, ATTO_LOG_TOPOL,
                  "esashba_get_bay_identifier: unknown topology for host %d\n",
                  shost->host_no);
        return -ENXIO;
    }
    ATTO_LOG (KERN_DEBUG, ATTO_LOG_TOPOL,
              "esashba_get_bay_identifier: found topology for host%d\n",
              shost->host_no);

    edev = esas_device_by_address (etop, rphy->identify.sas_address);
    if (!edev)
    {
        ATTO_LOG (KERN_WARN, ATTO_LOG_TOPOL,
                  "esashba_get_bay_identifier: unknown device for rphy 0x%llx\n",
                  rphy->identify.sas_address);
        return -ENXIO;
    }
    ATTO_LOG (KERN_DEBUG, ATTO_LOG_TOPOL,
              "esashba_get_bay_identifier: found device at HW target %d for rphy\n", edev->target_id);

    bay = OsTransport (shost->host_no, ATTO_TRANSPORT_DEVICE_GET_SLOT_NUM, edev->target_id, NULL);

    ATTO_LOG (KERN_INFO, ATTO_LOG_TOPOL, "esashba_get_bay_identifier: bay %d found for rphy 0x%llx\n",
              bay, rphy->identify.sas_address);

#endif /* ESAS2HBA || ESAS2HBA */
    return bay;
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
int esashba_set_phy_speed (struct sas_phy *phy, struct sas_phy_linkrates *rate)
{

    return -ENXIO;
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21)
int esashba_phy_enable (struct sas_phy *phy, int enable)
{

    return -ENXIO;
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
int esashba_smp_handler (struct Scsi_Host *sh, struct sas_rphy *rphy, struct request *req)
{
    int ret;

#ifdef ATTO_SAS_PORTS
    struct request *rsp = req->next_rq;
    int            type;
    ESAS_SMP_TASK  task;

    if (!rsp)
    {
        printk (DEBUG_HEADER "%s: no space for smp response\n", __func__);
        ret = -EINVAL;
        goto out;
    }

    if (!rphy)
    {
        printk (DEBUG_HEADER "%s: SMP commands only supported on remote phys.\n", __func__);
        ret = -EINVAL;
        goto out;
    }

    type = rphy->identify.device_type;
    if ((type != SAS_EDGE_EXPANDER_DEVICE) &&
        (type != SAS_FANOUT_EXPANDER_DEVICE))
    {
        printk (DEBUG_HEADER "%s: SMP commands only supported by expanders.\n", __func__);
        ret = -EINVAL;
        goto out;
    }

    task.SasAddress = rphy->identify.sas_address;
    task.Request = bio_data(req->bio);
    task.Response = bio_data(rsp->bio);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
    task.RequestLen = blk_rq_bytes(req);
    task.ResponseLen = blk_rq_bytes(rsp);
#else
    task.RequestLen = req->data_len;
    task.ResponseLen = rsp->data_len;
#endif

    if (OsTransport (sh->host_no,
                     ATTO_TRANSPORT_SMP_TASK,
                     0,
                     &task))
    {
        printk (DEBUG_HEADER "%s: SMP request failed, result %d\n", __func__, task.Result);
        ret = -ENXIO;
        goto out;
    }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31)
    req->resid_len = task.RequestLen;
    rsp->resid_len = task.ResponseLen;
#else
    req->data_len = task.RequestLen;
    rsp->data_len = task.ResponseLen;
#endif

    ret = 0;

#else
    ret = -ENXIO;
#endif

out:
    return ret;

}
#endif

/* The sas_function_template defines the entry points for             */
/* the SAS Transport API. These are based kernel version availability */
struct sas_function_template esashba_transport_functions =
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15)
    .get_linkerrors           = esashba_get_linkerrors,
    .phy_reset                = esashba_phy_reset,
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,18)
    .get_bay_identifier       = esashba_get_bay_identifier,
    .get_enclosure_identifier = NULL,
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
    .set_phy_speed            = esashba_set_phy_speed,
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,21)
    .phy_enable               = esashba_phy_enable,
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,23)
    .smp_handler              = esashba_smp_handler,
#endif

/* VMware requires this entry point */
#ifdef __VMKLNX__
    .get_target_sas_identifier = esashba_get_target_sas_identifier,
#endif
};


static struct scsi_transport_template *esashba_transport_template = NULL;


/****************************************************************************

    FUNCTION:   atto_get_transport_template

    PURPOSE:    Get the transport template for our driver

    ARGUMENTS:  None

    RETURNS:    Pointer to our transport template

    NOTES:      The transport template is created in atto_attach_transport

****************************************************************************/

struct scsi_transport_template *atto_get_transport_template (void)
{
    return esashba_transport_template;
}


/****************************************************************************

    FUNCTION:   atto_attach_transport

    PURPOSE:    Set up the transport template by attaching our entry points

    ARGUMENTS:  None

    RETURNS:    1 on success, 0 on failure

****************************************************************************/

int atto_attach_transport (void)
{
    int i;

    ATTO_LOG (KERN_INFO, ATTO_LOG_XPORT, "sas_attach_transport() called.\n");

    esashba_transport_template = sas_attach_transport (&esashba_transport_functions);

    if (esashba_transport_template == NULL)
    {
        printk (DEBUG_HEADER "sas_attach_transport() failed.");
        return 0;
    }

    for (i = 0; i < 32; ++i)
       esas_topology[i].host_no = -1;

    return 1;
}

void atto_release_transport (void)
{
    ATTO_LOG (KERN_INFO, ATTO_LOG_XPORT, "atto_release_transport()\n");

    if (esashba_transport_template)
    {
        ATTO_LOG (KERN_INFO, ATTO_LOG_XPORT, "sas_release_transport() called.\n");
        sas_release_transport (esashba_transport_template);
        esashba_transport_template = NULL;
    }

#ifdef ATTO_SAS_PORTS
    {
        int i;

        for (i = 0; i < 32; ++i)
        {
            if (esas_topology[i].status != ET_STATUS_NONE)
            {
                mutex_lock (&esas_topology[i].sas_topology_mutex);

                /* By the time we get here, host_device->dev is no longer valid */
                /* and may in fact be a pointer to a bunch of junk.  Make sure  */
                /* we don't try to use it by clearing it out now.               */

                esas_topology[i].host_device->dev = NULL;

                ATTO_LOG (KERN_INFO, ATTO_LOG_TOPOL,
                          "Destroying SAS topology (%d).\n", i);

                esas_topology_destroy (&esas_topology[i]);
                mutex_unlock (&esas_topology[i].sas_topology_mutex);
            }
        }
    }
#endif
}


void atto_init_host_attr (struct Scsi_Host *sh, int index)
{
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_XPORT, &sh->shost_gendev,
                  "atto_init_host_attr(%p, %d)\n", sh, index);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
    if (sh)
    {

#ifdef ATTO_SAS_PORTS
        {
            PESASTOPOLOGY etop = &esas_topology[index];
            u64           sas_address;
            PESASDEVICE   ehost;
            u16           i;

            ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, &sh->shost_gendev,
                          "Initializing SAS topology\n");
            /* Initialize sas_device_list & mutex */
            etop->status = ET_STATUS_INIT;
            INIT_LIST_HEAD (&etop->sas_device_list);
            mutex_init (&etop->sas_topology_mutex);
            etop->host_no = sh->host_no;
            etop->phy_index = 0;

            /* Create the host device and add it to the topology */

            sas_address = OsTransport (etop->host_no,
                                       ATTO_TRANSPORT_HOST_SAS_ADDRESS,
                                       0, NULL);

            ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, &sh->shost_gendev,
                          "Adding SAS Host Device (0x%llx) to topology for adapter %d.\n",
                          sas_address, index);

            ehost = esas_device_create (sas_address);
            if (!ehost)
            {
                /* out of memory */
                printk (DEBUG_HEADER "Failed to allocate SAS Host Device\n");
                return;
            }

            /* Set host attributes */
            ehost->flags = EDF_IS_HOST;
            ehost->dev = &sh->shost_gendev;

            /* Add phys to host */

            ehost->num_phys = (u16)OsTransport (etop->host_no,
                                                ATTO_TRANSPORT_HOST_NUM_PHYS,
                                                0, NULL);

            ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, ehost->dev,
                          "Adding %d phys to SAS Host\n", ehost->num_phys);

            ehost->phy_list = (struct sas_phy **) vmalloc (ehost->num_phys * sizeof (struct sas_phy *));

            if (!ehost->phy_list)
            {
                /* out of memory */
                printk (DEBUG_HEADER "Failed to allocate PHY list for SAS Host %d\n", index);
                return;
            }

            for (i = 0; i < ehost->num_phys; ++i)
            {
                struct sas_phy *phy;
                ESAS_SMP_DISCOVER hostphy;

                memset (&hostphy, 0, sizeof(ESAS_SMP_DISCOVER));

                phy = sas_phy_alloc (ehost->dev, etop->phy_index++);

                if (!phy)
                {
                    /* out of memory */
                    printk (DEBUG_HEADER "sas_phy_alloc(%p, %d) failed.\n",
                            ehost->dev, i);

                    ehost->phy_list[i] = NULL;
                    continue;
                }

                ehost->phy_list[i] = phy;

                /* Set default values as though the PHY is not used. */

                phy->identify.phy_identifier = i;
                phy->identify.device_type    = SAS_PHY_UNUSED;
                phy->negotiated_linkrate     =
                phy->maximum_linkrate_hw     =
                phy->maximum_linkrate        =
                phy->minimum_linkrate_hw     =
                phy->minimum_linkrate        = SAS_LINK_RATE_UNKNOWN;

                if (OsTransport (etop->host_no,
                                 ATTO_TRANSPORT_HOST_PHY_DISCOVER,
                                 i,
                                 &hostphy))
                {
                    /* report the failure, but add the phy */
                    printk (DEBUG_HEADER "PHY Discover failed for phy 0x%02x\n",i);
                }
                else
                {
                    /* Set phy attributes */
                    phy->identify.device_type = SAS_END_DEVICE;
                    phy->identify.initiator_port_protocols =
                        atto_proto_to_sas_proto (hostphy.AttachedInitiatorProto,0);
                    phy->identify.target_port_protocols =
                        atto_proto_to_sas_proto (hostphy.AttachedTargetProto, 1);
                    phy->identify.sas_address = ehost->sas_address;
                    
                    phy->negotiated_linkrate = 
                        atto_linkrate_to_sas_linkrate (hostphy.NegotiatedLogicalLinkRate);
                    phy->maximum_linkrate_hw = 
                        atto_linkrate_to_sas_linkrate (hostphy.HardwareMaxLinkRate);
                    phy->maximum_linkrate    =
                        atto_linkrate_to_sas_linkrate (hostphy.ProgrammedMaxLinkRate);
                    phy->minimum_linkrate_hw =
                        atto_linkrate_to_sas_linkrate (hostphy.HardwareMinLinkRate);
                    phy->minimum_linkrate    =
                        atto_linkrate_to_sas_linkrate (hostphy.ProgrammedMinLinkRate);
                }

                ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, ehost->dev,
                              "sas_phy_add() called.\n");
                if (sas_phy_add (phy))
                {
                    /* Failed to add phy */
                    printk (DEBUG_HEADER "sas_phy_add(%d) failed.\n", i);

                    sas_phy_free (phy);
                    ehost->phy_list[i] = NULL;
                    continue;
                }
            }

            INIT_LIST_HEAD (&ehost->port_list);

            /* Store the host in the topology */
            etop->host_device = ehost;
            etop->status = ET_STATUS_RUNNING;

            ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, ehost->dev,
                          "SAS Topology created for host: %d phys  0x%llx\n",
                          ehost->num_phys, ehost->sas_address);
        }

#endif

    }
    else
        printk (DEBUG_HEADER "atto_init_host_attr got NULL ptr (%p, %p)\n",
                sh, esashba_transport_template);
#endif
}


#ifdef ATTO_SAS_PORTS

/****************************************************************************

    FUNCTION:   esas_topology_add_device (PESASTOPOLOGY etopology,
                                          u16 target_id, u8 device_type)

    PURPOSE:    Add a device with the type and target id

    ARGUMENTS:  etopology - the topology for the new address
                target_id - The HW target ID of the device
                device_type - The type of the device

    RETURNS:    None

    NOTES:      

****************************************************************************/

void esas_topology_add_device (PESASTOPOLOGY etopology,
                               u16 target_id, u8 device_type)
{
    u64               sas_address;
    u16               parent_target_id;
    PESASDEVICE       parent_device;
    PESASDEVICE       this_device;
    PESASPORT         eport;
    u8                direct_attached;
    char              *ps;
    const char        *name;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etopology->host_device->dev,
                  "Add device at target 0x%04x\n", target_id);

    /* Using SAS topology and it is active? */
    if (etopology->status != ET_STATUS_RUNNING)
    {
        printk (DEBUG_HEADER "topology not running\n");
        return;
    }

    /* Sanity check for existing device with given target id */
    this_device = esas_device_by_target_id (etopology, target_id);
    if (this_device)
    {
        printk (DEBUG_HEADER "add device failure - duplicate target id  0x%04x\n",                target_id);
                return;
    }

    direct_attached = (u8) OsTransport (etopology->host_no,
                                        ATTO_TRANSPORT_DEVICE_IS_DIRECT_ATTACHED,
                                        target_id, NULL);
    if (direct_attached)
        parent_device = etopology->host_device;
    else
    {
        /* make sure the parent device is in the topology */
        parent_target_id = (u16) OsTransport (etopology->host_no,
                                              ATTO_TRANSPORT_DEVICE_GET_ATTACHED_TARGET_ID,
                                              target_id, NULL);
        parent_device = esas_device_by_target_id (etopology,
                                                  parent_target_id);
        if (!parent_device)
        {
            printk (DEBUG_HEADER "add device failure - parent not found\n");
            return;
        }
    }

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, parent_device->dev,
                  "parent address: 0x%llx\n",
                  parent_device->sas_address);

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, parent_device->dev,
                  "parent flags: 0x%x\n", parent_device->flags);

    sas_address = OsTransport (etopology->host_no,
                               ATTO_TRANSPORT_DEVICE_GET_SAS_ADDR,
                               target_id, NULL);

    /* Create a new device */
    this_device = esas_device_create (sas_address);

    if (this_device == NULL)
    {
        printk (DEBUG_HEADER "add device failure - cannot create device at target 0x%04x\n", target_id);
        return;
    }

    /* Set the device attributes */
    this_device->target_id = target_id;
    this_device->parent_device = parent_device;

    /* Create a new port to link between parent device and this device */
    eport = esas_port_create();
    if (!eport)
    {
        printk (DEBUG_HEADER "add device failure - cannot allocate port 0x%llx, 0x%llx.\n", parent_device->sas_address, sas_address);
        goto error_device;
    }

    /* Allocate a SAS Transport port */
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, parent_device->dev,
                  "sas_port_alloc_num() called.\n");
    eport->port = sas_port_alloc_num (parent_device->dev);
    if (!eport->port)
    {
        printk (DEBUG_HEADER "sas_port_alloc_num failed on 0x%llx.\n",
                parent_device->sas_address);
        goto error_port;
    }

    /* Put the SAS Transport port into the mid-layer */
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, parent_device->dev,
                  "sas_port_add() called.\n");
    if (sas_port_add (eport->port))
    {
        printk (DEBUG_HEADER "sas_port_add failed on 0x%llx\n",
                parent_device->sas_address);
#ifndef __VMKLNX__
        sas_port_free (eport->port);
#endif
        goto error_port;
    }

    /* The port points to both devices. The near device is the one closer */
    /* to the host.                                                       */
    eport->near_device = parent_device;
    eport->far_device = this_device;

    /* Add phys from parent device to the port. */

    if (parent_device->flags & EDF_IS_HOST)
    {
        /* If the parent device is the host, put host phys in the port    */
        /* Ignore the host phys result if the device is virtual             */
        
        if ( (esas_port_add_host_phys (etopology, eport) == 0)
              && (!OsTransport(etopology->host_no,
                               ATTO_TRANSPORT_DEVICE_IS_VIRTUAL,
                               target_id, NULL)) )
        {
            printk (DEBUG_HEADER "error adding host phys from 0x%llx to port\n", parent_device->sas_address);
            goto error_added_port;
        }
    }
    else if (parent_device->flags & EDF_IS_EXPANDER)
    {
        /* If the parent device is an expander put those phys in the port */
        if (esas_port_add_expander_phys (etopology, eport) == 0)
        {
            printk (DEBUG_HEADER "error adding phys from 0x%llx to port\n",
                    parent_device->sas_address);
            goto error_added_port;
        }
    }
    else
    {
        /* What the heck is it? */
        printk (DEBUG_HEADER "unknown parent device type (0x%x) at 0x%llx\n",
                parent_device->flags, parent_device->sas_address);
        goto error_added_port;
    }

    /* The port goes on the parent device */
    list_add_tail (&eport->list, &parent_device->port_list);

    /* Put this device in the topology before making an end-device */
    /* or expander device, as the topology may be searched during  */
    /* the next phase.                                             */
    list_add_tail (&this_device->list, &etopology->sas_device_list);

    if (device_type == ATTO_DEVICE_TYPE_EXPANDER)
    {
        if (esas_device_make_expander (etopology, eport, this_device))
        {
            printk (DEBUG_HEADER "error creating expander for device 0x%llx\n", this_device->sas_address);
            goto error_added_device;
        }
        if (esas_device_add_parent_port (etopology, this_device))
        {
            printk (DEBUG_HEADER "error adding parent port to 0x%llx.\n", this_device->sas_address);
            /* This is not a hard failure condition; keep going */
        }
    }
    else if (device_type == ATTO_DEVICE_TYPE_TARGET)
    {
        if (esas_device_make_end_device (etopology, eport, this_device))
        {
            printk (DEBUG_HEADER "erro creating end device for device 0x%llx\n", this_device->sas_address);
            goto error_added_device;
        }

    }
    else
    {
        printk (DEBUG_HEADER "Unknown device type %d for device 0x%llx\n", device_type, sas_address);
        goto error_added_device;
    }

    if (this_device->rphy->identify.target_port_protocols & SAS_PROTOCOL_SATA)
        ps = "sata";
    if (this_device->rphy->identify.target_port_protocols & SAS_PROTOCOL_SMP)
        ps = "smp";
    if (this_device->rphy->identify.target_port_protocols & SAS_PROTOCOL_STP)
        ps = "stp";
    if (this_device->rphy->identify.target_port_protocols & SAS_PROTOCOL_SSP)
        ps = "ssp";
    

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, this_device->dev,
                  "added device 0x%llx\n",
                  this_device->sas_address);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,30)
    name = dev_name(&this_device->rphy->dev);
#else
    name = this_device->rphy->dev.bus_id;
#endif

    esas_device_set_active(this_device);

    printk (DEBUG_HEADER "added %s device, %s\n",
            ps, name);

    return;

error_added_device:
    list_del (&this_device->list);
    list_del (&eport->list);

error_added_port:
    sas_port_delete (eport->port);

error_port:
    esas_port_destroy (eport);

error_device:
    esas_device_destroy (this_device);

    return;
}

/****************************************************************************

    FUNCTION:   esas_topology_remove_device (PESASTOPOLOGY etopology,
                                             u16 target_id)

    PURPOSE:    Remove the device at the given target id from the topology

    ARGUMENTS:  etopology - the topology containing the address
                target_id - The target ID of the device

    RETURNS:    

****************************************************************************/

void esas_topology_remove_device (PESASTOPOLOGY etopology,
                                  u16 target_id)
{
    PESASDEVICE this_device;
    PESASDEVICE parent_device;
    PESASPORT   eport;
    char        *ps;
    const char  *name;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etopology->host_device->dev,
                  "Remove device at 0x%04x\n", target_id);

    /* Find this device in topology. Not found is possible and is not */
    /* a hard error condition.                                        */
    this_device = esas_device_by_target_id (etopology, target_id);
    if (!this_device)
    {
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etopology->host_device->dev,
                      "Failed to find device for removal at 0x%04x\n",
                      target_id);
        return;
    }

    if (this_device->flags & EDF_IS_EXPANDER) 
    {
        PESASPORT p, tp;

        /* remove sub-tree beneath this expander */
        list_for_each_entry_safe (p, tp, &this_device->port_list, list)
        {
            esas_topology_remove_device (etopology,
                                         p->far_device->target_id);
        }

    }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,30)
    name = dev_name(&this_device->rphy->dev);
#else
    name = this_device->rphy->dev.bus_id;
#endif

    if (this_device->rphy->identify.target_port_protocols & SAS_PROTOCOL_SATA)
        ps = "sata";
    if (this_device->rphy->identify.target_port_protocols & SAS_PROTOCOL_SMP)
        ps = "smp";
    if (this_device->rphy->identify.target_port_protocols & SAS_PROTOCOL_STP)
        ps = "stp";
    if (this_device->rphy->identify.target_port_protocols & SAS_PROTOCOL_SSP)
        ps = "ssp";


    /* Log this now, as the device name will go away when the device */
    /* is removed.                                                   */
    printk (DEBUG_HEADER "removed %s device, %s\n",
            ps, name);

    
    /* Save this device's parent device pointer */
    parent_device = this_device->parent_device;

    /* Find the eport from the parent device's portlist */
    eport = esas_port_by_sas_address (parent_device,
                                      this_device->sas_address);
    if (!eport)
    {
        printk (DEBUG_HEADER "error removing device - could not find port for device 0x%llx\n", parent_device->sas_address);
        return;
    }

    /* If this device is an expander, remove the parent port */
    /* and free the phy_list                                 */
    if (this_device->flags & EDF_IS_EXPANDER)
    {
        esas_device_remove_parent_port (this_device);
        vfree (this_device->phy_list);
    }

    /* Delete the sas port; sas_port_delete() deletes the end device (rphy) */
    /* and unlinks the sas phys from the port. It also deletes the phys     */
    /* on the rphy instance. */
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, parent_device->dev,
                  "sas_port_delete() called.\n");
    sas_port_delete (eport->port);

    /* remove the eport from the parent device's portlist */
    list_del (&eport->list);

    /* Free the eport */
    esas_port_destroy (eport);

    /* Remove this device from topology */
    list_del (&this_device->list);

    /* Free this device */
    esas_device_destroy (this_device);

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etopology->host_device->dev,
                  "Device removed at target 0x%04x\n", target_id);

    return;
}


/****************************************************************************

    FUNCTION:   esas_topology_deactivate_device (PESASTOPOLOGY etopology,
                                                 u16 target_id)

    PURPOSE:    Deactivate the target with the given target ID

    ARGUMENTS:  etopology - the topology containing the address
                target_id - the target ID of the device

    RETURNS:    None

****************************************************************************/

void esas_topology_deactivate_device (PESASTOPOLOGY etopology,
                                      u16 target_id)
{
    PESASDEVICE device;

    device = esas_device_by_target_id (etopology, target_id);

    if (!device)
    {
        printk (DEBUG_HEADER "Failed to find device at target 0x%04d for deactivation\n",
                target_id);
        return;
    }

    esas_device_set_inactive(device);

}


/****************************************************************************

    FUNCTION:   esas_topology_activate_device (PESASTOPOLOGY etopology,
                                               u16 target_id)

    PURPOSE:    Activate the target with the given target ID

    ARGUMENTS:  etopology - the topology containing the target
                target_id - the target ID of the device

    RETURNS:    None

****************************************************************************/

void esas_topology_activate_device (PESASTOPOLOGY etopology,
                                    u16 target_id)
{
    PESASDEVICE device;

    device = esas_device_by_target_id (etopology, target_id);

    if (!device)
    {
        printk (DEBUG_HEADER "Failed to find device at target 0x%04d for activation\n",
                target_id);
        return;
    }

    esas_device_set_active(device);
}


/****************************************************************************

    FUNCTION:   esas_topology_scan_device (PESASTOPOLOGY etopology,
                                           u16 target_id)

    PURPOSE:    Re-scan the target at the given SAS address

    ARGUMENTS:  etopology - the topology containing the address
                target_id - the target ID of the device

    RETURNS:    None

****************************************************************************/

void esas_topology_scan_device (PESASTOPOLOGY etopology,
                                u16 target_id)
{
    PESASDEVICE edevice;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etopology->host_device->dev,
                  "Scan device at target %d\n", target_id);

    edevice = esas_device_by_target_id (etopology, target_id);

    if (!edevice)
    {
        printk (DEBUG_HEADER "Failed to find device at target 0x%04d for rescan.\n",
                target_id);
        return;
    }

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, edevice->dev,
                  "scanning device\n");

    if (!(edevice->flags & EDF_IS_TARGET))
    {
        printk (DEBUG_HEADER "Attempt to invoke rescan on non-target device at target 0x%04x.\n", target_id);
        return;
    }

    if (!edevice->rphy)
    {
        printk (DEBUG_HEADER "No rphy for device at target 0x%04x rescan.\n",
                target_id);
        return;
    }

    /* Call mid-layer to invoke a scan on the port */
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_SCSI, edevice->dev,
                  "scsi_scan_target() called.\n");
    scsi_scan_target (edevice->dev, 0, edevice->rphy->scsi_target_id,
                      SCAN_WILD_CARD, 1);


    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etopology->host_device->dev,
                  "Scan device done\n");
}


/****************************************************************************

    FUNCTION:   esas_topology_validate_device (PESASTOPOLOGY etopology,
                                               u16 target_id)

    PURPOSE:    Validate the parent port phys of the device

    ARGUMENTS:  etopology - the topology containing the address
                target_id - the target ID of the device

    RETURNS:    None

****************************************************************************/

void esas_topology_validate_device (PESASTOPOLOGY etopology,
                                    u16 target_id)
{
    PESASDEVICE device, parent_device;
    PESASPORT   eport;
    struct sas_phy *phy, *tmp_phy;
    ESAS_SMP_DISCOVER sd;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etopology->host_device->dev,
                  "Validate device at target %d\n", target_id);

    device = esas_device_by_target_id (etopology, target_id);

    if (!device)
    {
        printk (DEBUG_HEADER "Failed to find device at target 0x%04d for validation.\n",
                target_id);
        return;
    }

    parent_device = device->parent_device;

    eport = esas_port_by_sas_address (device->parent_device,
                                      device->sas_address);

    mutex_lock (&eport->port->phy_list_mutex);
    list_for_each_entry_safe (phy, tmp_phy, &eport->port->phy_list,
                              port_siblings)
    {
        /* validate each phy in port */
        memset (&sd, 0, sizeof (ESAS_SMP_DISCOVER));
        sd.PhyId = phy->identify.phy_identifier;
        sd.SasAddress = parent_device->sas_address;

        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, parent_device->dev,
                      "Validate phy 0x%x on 0x%llx\n",
                      sd.PhyId, sd.SasAddress);

        if (parent_device->flags & EDF_IS_HOST)
        {
            if (OsTransport (etopology->host_no,
                             ATTO_TRANSPORT_HOST_PHY_DISCOVER,
                             sd.PhyId,
                             &sd))
            {
                printk (DEBUG_HEADER "Host PHY Discover failed for PHY 0x%x on 0x%llx.\n", sd.PhyId, parent_device->sas_address);
                continue;
            }
        }
        else if (parent_device->flags & EDF_IS_EXPANDER)
        {
            if (OsTransport (etopology->host_no,
                             ATTO_TRANSPORT_SMP_DISCOVER,
                             parent_device->target_id,
                             &sd))
            {
                printk (DEBUG_HEADER "SMP Discover failed for PHY 0x%x on 0x%llx.\n", sd.PhyId, parent_device->sas_address);
                continue;
            }

        }
        if (atto_linkrate_to_sas_linkrate (sd.NegotiatedLogicalLinkRate) == SAS_LINK_RATE_UNKNOWN)
        {
            ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, parent_device->dev,
                          "Negotiated linkrate unknown for phy 0x%x on 0x%llx\n",
                          sd.PhyId, sd.SasAddress);
            continue;
        }

        phy->negotiated_linkrate = atto_linkrate_to_sas_linkrate (sd.NegotiatedLogicalLinkRate);
        
    }
    mutex_unlock (&eport->port->phy_list_mutex);
}


/****************************************************************************

    FUNCTION:   esas_topology_list_devices (PESASTOPOLOGY etopology char *buffer))

    PURPOSE:    List the devices in the topology in the order they were added
                The format of this function needs to match the output
                from the proc ioctl that lists devices. This is also
                designed specifically for Hitachi.

    ARGUMENTS:  t - An adapter's SAS topology
                buffer - scratch buffer to contain device info

    RETURNS:    None

****************************************************************************/

int esas_topology_list_devices (PESASTOPOLOGY etopology, char *buffer)
{
    PESASDEVICE d, td;
    int count = 1, size = 0;

    if (!etopology)
        return 0;

    /* List targets ... */
    list_for_each_entry_safe (d, td, &etopology->sas_device_list, list)
    {
        if (d->flags & EDF_IS_TARGET)
        {
            size += sprintf (buffer + size, "%4d  %4d   %08X:%08X  %4d\n",
                             count,
                             count-1,
                             (u32)(d->sas_address >> 32),
                             (u32)d->sas_address,
                             d->target_id);
            ++count;
        }
    }

    return size;
}

extern int num_targets;


/****************************************************************************

    FUNCTION:   esas_topology_get_device_relative_order
                           (PESASTOPOLOGY etopology, int target_id))

    PURPOSE:    Get the relative ordering (0-based integer) of the device
                specified by the given HW target id.  Only consider
                target devices when determining the relative order.

    ARGUMENTS:  etopology - An adapter's SAS topology
                target_id - A devices HW-layer target id.

    RETURNS:    The relative ordering as a 0-based integer, or num_targets
                if the device cannot be found.

****************************************************************************/

int esas_topology_get_device_relative_order (PESASTOPOLOGY etopology, int target_id)
{
   PESASDEVICE d, td;
   int position = 0, found = 0;

    if (!etopology)
        return num_targets;

    /* Walk target list until found ... */
    list_for_each_entry_safe (d, td, &etopology->sas_device_list, list)
    {
        if (d->flags & EDF_IS_TARGET)
        {
            if (d->target_id == target_id)
            {
                found = 1;
                break;
            }
            ++position;
        }
    }

    if (!found)
        return num_targets;
    else
        return position;
}



/****************************************************************************

    FUNCTION:   esas_topology_destroy (PESASTOPOLOGY etopology)

    PURPOSE:    Destroy the topology - this is just an exercise in
                freeing all the memory associated with the devices/ports.

    ARGUMENTS:  t - An adapter's SAS topology

    RETURNS:    None

****************************************************************************/

void esas_topology_destroy (PESASTOPOLOGY etopology)
{
    PESASDEVICE d, td, ehost;
    PESASPORT p, tp;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etopology->host_device->dev,
                  "Destroy topology\n");

    /* Force all devices to be removed */
    etopology->status = ET_STATUS_UNLOADING;

    /* safely traverse the topology - for each device that is a target */
    /* call esas_topology_remove_device(device).                       */

    list_for_each_entry_safe (d, td, &etopology->sas_device_list, list)
    {
        /* Be careful, d->dev may no longer be valid. */

        ATTO_LOG (KERN_INFO, ATTO_LOG_TOPOL,
                  "destroying device 0x%llx\n", d->sas_address);
        if (d->flags & EDF_IS_EXPANDER)
        {
            list_for_each_entry_safe (p, tp, &d->port_list, list)
            {
                ATTO_LOG (KERN_INFO, ATTO_LOG_TOPOL,
                          "destroying port %p\n", p);
                list_del (&p->list);
                esas_port_destroy (p);
            }

            vfree (d->phy_list);
            d->phy_list = NULL;

            esas_device_remove_parent_port (d);
        }

        list_del (&d->list);

        esas_device_destroy (d);
    }

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etopology->host_device->dev,
                  "destroying host device\n");

    /* Clean up the etopology->host */
    ehost = etopology->host_device;

    /*   free phy_list       */
    vfree (ehost->phy_list);
    ehost->phy_list = NULL;

    /*   free the host device */
    esas_device_destroy (ehost);

    etopology->host_device = NULL;

    etopology->status = ET_STATUS_NONE;

    ATTO_LOG (KERN_INFO, ATTO_LOG_TOPOL, "Topology destroy complete.\n");
}


/****************************************************************************

    FUNCTION:   esas_topology_by_host_no (u32 host_no)

    PURPOSE:    Find the ESAS Topology for a given host number

    ARGUMENTS:  host_no - The linux scsi host number

    RETURNS:    Pointer to ESAS Topology, NULL if not found.

****************************************************************************/

PESASTOPOLOGY esas_topology_by_host_no (u32 host_no)
{
    int i;
    
    for (i = 0; i < 32; ++i)
    {
        if (esas_topology[i].host_no == host_no)
            return &esas_topology[i];
    }

    return NULL;
}


/****************************************************************************

    FUNCTION:   esas_device_by_address (PESASTOPOLOGY t,
                                        u64 a)

    PURPOSE:    Find a device in the given topology with the given
                SAS address

    ARGUMENTS:  t - An adapter's SAS topology
                a - a device's SAS address

    RETURNS:    Pointer to device, if found; otherwise NULL

****************************************************************************/

PESASDEVICE esas_device_by_address (PESASTOPOLOGY t,
                                    u64 a)
{
    PESASDEVICE d;

    list_for_each_entry (d, &t->sas_device_list, list)
        if (d->sas_address == a)
            return d;

    return NULL;
}


/****************************************************************************

    FUNCTION:   esas_device_by_target_id (PESASTOPOLOGY t,
                                          u16 target_id)

    PURPOSE:    Find a device in the given topology with the given
                target id

    ARGUMENTS:  t - An adapter's SAS topology
                target_id - a device's target id

    RETURNS:    Pointer to device, if found; otherwise NULL

****************************************************************************/

PESASDEVICE esas_device_by_target_id (PESASTOPOLOGY t,
                                      u16 target_id)
{
    PESASDEVICE d;

    list_for_each_entry (d, &t->sas_device_list, list)
        if (d->target_id == target_id)
            return d;

    return NULL;
}


/****************************************************************************

    FUNCTION:   esas_device_create (u64 sas_address)

    PURPOSE:    Create an esas device object

    ARGUMENTS:  sas_address - the address of the new device

    RETURNS:    The newly created esas device object.

****************************************************************************/

PESASDEVICE esas_device_create (u64 sas_address)
{
    PESASDEVICE d;

    d = (PESASDEVICE) vmalloc (sizeof (ESASDEVICE));
    if (d == NULL)
    {
        printk (DEBUG_HEADER "failed to allocate ESASDEVICE\n");
        return d;
    }

    memset (d, 0, sizeof (ESASDEVICE));

    d->sas_address = sas_address;
    
    return d;
}


/****************************************************************************

    FUNCTION:   esas_device_destroy (PESASDEVICE edevice)

    PURPOSE:    Destroy an esas device object

    ARGUMENTS:  edevice - the device object to destroy

    RETURNS:    NONE

****************************************************************************/

void esas_device_destroy (PESASDEVICE edevice)
{
    if (edevice)
        vfree (edevice);
}


/****************************************************************************

    FUNCTION:   esas_device_make_expander (PESASTOPOLOGY etop,
                                           PESASPORT eport,
                                           PESASDEVICE edevice)

    PURPOSE:    Make an ESAS device into an ESAS expander device

    ARGUMENTS:  etop - The ESAS topology where the device reside
                eport - The ESAS port connected to this device
                edevice - the device to make into an expander

    RETURNS:    0 on success, 1 on failure

****************************************************************************/

int esas_device_make_expander (PESASTOPOLOGY etop, PESASPORT eport,
                               PESASDEVICE this_device)
{
    ESAS_EXPANDER_INFO         sei;
    struct sas_expander_device *sed;
    ESAS_SMP_DISCOVER          near_phy_info;
    int                        near_phy_id;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etop->host_device->dev,
                  "Make expander at target %d\n", this_device->target_id);

    INIT_LIST_HEAD (&this_device->port_list);

    /* Get some info from the PHY that discovered this device */
    if ((near_phy_id = esas_port_get_near_phy_info (etop, eport, &near_phy_info)) < 0)
    {
        printk (DEBUG_HEADER "Failed to get near PHY info for device 0x%llx\n",
                this_device->sas_address);
        return 1;
    }

    this_device->flags = EDF_IS_EDGE_EXPANDER;

    /* Alloc expander device from the port */
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etop->host_device->dev,
                  "sas_expander_alloc() called.\n");
    this_device->rphy = sas_expander_alloc (eport->port, SAS_EDGE_EXPANDER_DEVICE);
    if (!this_device->rphy)
    {
        printk (DEBUG_HEADER "sas_expander_alloc failed for device 0x%llx\n",
                eport->near_device->sas_address);
        return 1;
    }

    /* Get a pointer back to the SAS expander device */
    sed = rphy_to_expander_device (this_device->rphy);
    if (!sed)
    {
        printk (DEBUG_HEADER "rphy_to_expander_device failed for device 0x%llx\n", eport->near_device->sas_address);
        sas_rphy_free (this_device->rphy);
        return 1;
    }

    /* Set rphy attributes */
    this_device->rphy->identify.target_port_protocols = SAS_PROTOCOL_SMP;
    this_device->rphy->identify.initiator_port_protocols = SAS_PROTOCOL_SMP;
    this_device->rphy->identify.sas_address = this_device->sas_address;
    this_device->rphy->identify.phy_identifier = (u8)near_phy_id;

    /* Set expander attributes: vendor, component, etc            */
    /* NOTE: For 3Gb, this information is not available and will  */
    /* be reported as emptry strings. This is a known limitation. */
    if (!OsTransport (etop->host_no,
                      ATTO_TRANSPORT_DEVICE_GET_EXP_INFO,
                      this_device->target_id, &sei))
    {
        memcpy (sed->vendor_id, sei.VendorId, SAS_EXPANDER_VENDOR_ID_LEN);
        memcpy (sed->product_id, sei.ProductId, SAS_EXPANDER_PRODUCT_ID_LEN);
        memcpy (sed->product_rev, sei.Revision, SAS_EXPANDER_PRODUCT_REV_LEN);
        memcpy (sed->component_vendor_id, sei.ComponentVendor,
                SAS_EXPANDER_COMPONENT_VENDOR_ID_LEN);
        sed->component_id = sei.ComponentId;
        sed->component_revision_id = sei.ComponentRev;
    }

    /* Add rphy */
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etop->host_device->dev,
                  "sas_rphy_add() called.\n");
    if (sas_rphy_add (this_device->rphy))
    {
        sas_rphy_free (this_device->rphy);
        printk (DEBUG_HEADER "sas_rphy_add failed for device 0x%llx.\n",
                eport->near_device->sas_address);
        return 1;
    }

    this_device->dev = &this_device->rphy->dev;

    /* Add phys to this_device. */
    esas_device_add_expander_phys (etop, this_device);

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etop->host_device->dev,
                  "Make expander device done\n");

    return 0;
}


/****************************************************************************

    FUNCTION:   esas_device_make_end (PESASTOPOLOGY etop,
                                      PESASPORT eport,
                                      PESASDEVICE this_device)

    PURPOSE:    Make an ESAS device into an ESAS end device

    ARGUMENTS:  etop - The ESAS topology where the device reside
                eport - The ESAS port connected to this device
                this_device - the device to make into an end device

    RETURNS:    0 on success, 1 on failure

****************************************************************************/

int esas_device_make_end_device (PESASTOPOLOGY etop,
                                 PESASPORT eport,
                                 PESASDEVICE this_device)
{
    u8 target_port_proto;
    u64 protocols;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etop->host_device->dev,
                  "Make end device at target %d\n", this_device->target_id);

    this_device->flags = EDF_IS_TARGET;

    /* Alloc end device from the port */
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etop->host_device->dev,
                  "sas_end_device_alloc() called.\n");
    this_device->rphy = sas_end_device_alloc (eport->port);
    if (!this_device->rphy)
    {
        printk (DEBUG_HEADER "sas_end_device_alloc failed for device 0x%llx.\n", eport->near_device->sas_address);
        return 1;
    }

    /* Set rphy attributes */
    protocols = OsTransport (etop->host_no,
                             ATTO_TRANSPORT_DEVICE_GET_PROTO,
                             this_device->target_id, NULL);
    target_port_proto = atto_proto_to_sas_proto (protocols, 1);

    this_device->rphy->identify.target_port_protocols = target_port_proto;
    this_device->rphy->identify.sas_address = this_device->sas_address;
    this_device->rphy->identify.phy_identifier = 0;
    
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etop->host_device->dev,
                  "sas_rphy:target_port_protocols:0x%x\n",
                   this_device->rphy->identify.target_port_protocols);

    /* Add rphy */
    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etop->host_device->dev,
                  "sas_rphy_add() called.\n");
    if (sas_rphy_add (this_device->rphy))
    {
        printk (DEBUG_HEADER "sas_rphy_add failed for device 0x%llx.\n",
                this_device->sas_address);
        sas_rphy_free (this_device->rphy);
        return 1;
    }

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etop->host_device->dev,
                  "sas_rphy:scsi_target_id:0x%x\n",
                  this_device->rphy->scsi_target_id);

    this_device->dev = &this_device->rphy->dev;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, etop->host_device->dev,
                  "Make end device done\n");
    return 0;
}


/****************************************************************************

    FUNCTION:   esas_device_add_expander_phys (PESASDEVICE edevice)

    PURPOSE:    Add the phys to an expander device

    ARGUMENTS:  edevice - the expander device

    RETURNS:    Number of phys added; 0 indicates a failure

    NOTES:      This function uses SMP commands to obtain
                information about the expander and its PHYs

****************************************************************************/

int esas_device_add_expander_phys (PESASTOPOLOGY etop, PESASDEVICE edev)
{
    ESAS_SMP_REPORT_GENERAL srg;
    ESAS_SMP_DISCOVER       sd;
    int                     i;
    struct sas_phy          *phy = NULL;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, edev->dev,
                  "Add phys to expander at target %d\n", edev->target_id);

    /* Use SMP Report General to obtain the # of phys in the expander */
    srg.SasAddress = edev->sas_address;
    if (OsTransport (etop->host_no,
                     ATTO_TRANSPORT_SMP_REPORT_GENERAL,
                     edev->target_id,
                     &srg))
    {
        printk (DEBUG_HEADER "SMP Report General failure for device 0x%llx\n",
            edev->sas_address);
        return 0;
    }

    edev->num_phys = srg.NumPhys;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, edev->dev,
                  "found %d phys on expander\n", edev->num_phys);

    edev->phy_list = (struct sas_phy **) vmalloc (edev->num_phys * sizeof (struct sas_phy *));
    if (!edev->phy_list)
    {
        printk (DEBUG_HEADER "Failed to allocate PHY list for SAS Device 0x%llx\n",
                edev->sas_address);
        return 0;
    }

    /* Use SMP Discover to obtain info about each PHY */
    for (i = 0; i < edev->num_phys; ++i)
    {
        int retries = ATTO_PHY_DISCOVERY_RETRIES;

        memset (&sd, 0, sizeof (ESAS_SMP_DISCOVER));
        sd.PhyId = i;
        sd.SasAddress = edev->sas_address;

        phy = sas_phy_alloc (edev->dev, etop->phy_index++);

        if (!phy)
        {
            /* out of memory */
            printk (DEBUG_HEADER "sas_phy_alloc(%p, %d) failed.\n",
                    edev->dev, i);

            edev->phy_list[i] = NULL;
            continue;
        }

        edev->phy_list[i] = phy;

        /* Set default values as though the PHY is not used. */

        phy->identify.phy_identifier = i;
        phy->identify.device_type    = SAS_PHY_UNUSED;
        phy->negotiated_linkrate     =
        phy->maximum_linkrate_hw     =
        phy->maximum_linkrate        =
        phy->minimum_linkrate_hw     =
        phy->minimum_linkrate        = SAS_LINK_RATE_UNKNOWN;

RetryDiscovery:
        if (OsTransport(etop->host_no,
                        ATTO_TRANSPORT_SMP_DISCOVER,
                        edev->target_id,
                        &sd))
        {
            if (retries--)
            {
                ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, edev->dev,
                              "PHY 0x%02x SMP discovery retry %d on device 0x%llx\n",
                              i, retries, edev->sas_address);
                mdelay (ATTO_PHY_DISCOVERY_DELAY_MS);
                goto RetryDiscovery;
            }

            printk (DEBUG_HEADER "SMP Discover failed for PHY 0x%x on device 0x%llx\n", i, edev->sas_address);
            goto AddPhy;
        }

        if (atto_linkrate_to_sas_linkrate (sd.NegotiatedLogicalLinkRate)
        ==  SAS_LINK_RATE_UNKNOWN)
        {
            if (retries--)
            {
                ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, edev->dev,
                              "PHY 0x%02x SMP discovery retry %d\n",
                              i, retries);
                mdelay (ATTO_PHY_DISCOVERY_DELAY_MS);
                goto RetryDiscovery;
            }

            ATTO_DEV_LOG (KERN_WARN, ATTO_LOG_TOPOL, edev->dev,
                          "PHY link rate 0x%02x unknown for phy 0x%02x\n",
                          sd.NegotiatedLogicalLinkRate, i);

            goto AddPhy;
        }

        phy->identify.device_type = SAS_EDGE_EXPANDER_DEVICE;
        phy->identify.initiator_port_protocols = 
            atto_proto_to_sas_proto (sd.AttachedInitiatorProto, 0);
        phy->identify.target_port_protocols =
            atto_proto_to_sas_proto (sd.AttachedTargetProto, 1);
        phy->identify.sas_address = sd.SasAddress;
        phy->identify.phy_identifier = i;

        phy->number = i;
        phy->negotiated_linkrate =
            atto_linkrate_to_sas_linkrate (sd.NegotiatedLogicalLinkRate);
        phy->minimum_linkrate_hw =
            atto_linkrate_to_sas_linkrate (sd.HardwareMinLinkRate);
        phy->minimum_linkrate =
            atto_linkrate_to_sas_linkrate (sd.ProgrammedMinLinkRate);
        phy->maximum_linkrate_hw =
            atto_linkrate_to_sas_linkrate (sd.HardwareMaxLinkRate);
        phy->maximum_linkrate =
            atto_linkrate_to_sas_linkrate (sd.ProgrammedMaxLinkRate);

AddPhy:
        ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, edev->dev,
                      "sas_phy_add(%d) called.\n", i);
        if (sas_phy_add (phy))
        {
            printk (DEBUG_HEADER "sas_phy_add(%d) failed for device 0x%llx.\n",
                    i, edev->sas_address);

            sas_phy_free (phy);
            edev->phy_list[i] = NULL;
            continue;
        }
                
    }

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, edev->dev,
                  "Add expander phys done. Added %d phys.\n", edev->num_phys);

    return edev->num_phys;
}


/****************************************************************************

    FUNCTION:   esas_device_add_parent_port (PESASTOPOLOGY etop,
                                             PESASDEVICE edevice)

    PURPOSE:    Add a backlinked port from this device to it's parent

    ARGUMENTS:  etop - The SAS topology where this device resides
                this_device - the expander device

    RETURNS:    0 on success, 1 on error

    NOTES:      This function uses SMP commands to obtain
                information about the expander and its PHYs

****************************************************************************/

int esas_device_add_parent_port (PESASTOPOLOGY etop, PESASDEVICE this_device)
{

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, this_device->dev,
                  "Add parent port to target %d\n", this_device->target_id);

    this_device->parent_port = esas_port_create();
    if (!this_device->parent_port)
    {
        printk (DEBUG_HEADER "Failed to allocated esas port for device 0x%llx.\n", this_device->sas_address);
        return 1;
    }

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, this_device->dev,
                  "sas_port_alloc_num() called.\n");
    this_device->parent_port->port = sas_port_alloc_num (this_device->dev);
    if (!this_device->parent_port->port)
    {
        printk (DEBUG_HEADER "sas_port_alloc_num failed on device 0x%llx.\n",
                this_device->sas_address);
        return 1;
    }

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, this_device->dev,
                  "sas_port_add() called.\n");
    if (sas_port_add (this_device->parent_port->port))
    {
        printk (DEBUG_HEADER "sas_port_add failed on device 0x%llx.\n",
                this_device->sas_address);
#ifndef __VMKLNX__
        sas_port_free (this_device->parent_port->port);
#endif
        return 1;
    }

    this_device->parent_port->near_device = this_device;
    this_device->parent_port->far_device = this_device->parent_device;

    /* Don't put the parent port on the port list */

    /* Add expanders phys to the parent port */
    if (esas_port_add_expander_phys (etop, this_device->parent_port) == 0)
    {
        printk (DEBUG_HEADER "No expander phys from 0x%llx added to parent port.\n", this_device->sas_address);
        /* Leave this port alone, for now */
        return 1;
    }

    /* Indicate this is a backlinked port to the transport layer */
    sas_port_mark_backlink (this_device->parent_port->port);

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, this_device->dev,
                  "Add parent port done\n");

    return 0;
}


/****************************************************************************

    FUNCTION:   esas_device_remove_parent_port (PESASDEVICE this_device)

    PURPOSE:    Remove the backlinked port from this device to it's parent

    ARGUMENTS:  this_device - the expander device

    RETURNS:    0 on success, 1 on error

****************************************************************************/

void esas_device_remove_parent_port (PESASDEVICE this_device)
{
    /* Watch out, this_device->dev may no longer be valid. */
    ATTO_LOG (KERN_INFO, ATTO_LOG_TOPOL,
              "Remove parent port (%p).\n", this_device);
    /* No additional work needs to be done at the SAS transport layer */
    /* as the sas_port_delete call on this_device will clean up all   */
    /* children including the parent port.                            */
    if (this_device->parent_port)
        esas_port_destroy(this_device->parent_port);
    this_device->parent_port = NULL;

}


/****************************************************************************

    FUNCTION:   esas_device_is_valid (PESASTOPOLOGY etop, PESASDEVICE edev)

    PURPOSE:    Check whether a device is still valid

    ARGUMENTS:  etop - The topology where the device resides
                edev - The device

    RETURNS:    TRUE if device is still valid; FALSE o/w

****************************************************************************/

u8 esas_device_is_valid (PESASTOPOLOGY etop, PESASDEVICE edev)
{
    /* Check topology status for unloading */
    if (etop->status == ET_STATUS_UNLOADING)
        return 0;

    /* Check device is active */
    if (!esas_device_is_active(edev))
        return 0;

    /* TODO: Try to get sas address; if it succeeds device is still valid */
    if (OsTransport (etop->host_no,
                     ATTO_TRANSPORT_DEVICE_GET_SAS_ADDR,
                     edev->target_id,
                     NULL))
        return 1;
    else
        return 0;
}


/****************************************************************************

    FUNCTION:   esas_device_is_active (PESASDEVICE edev)

    PURPOSE:    Check whether a device is active

    ARGUMENTS:  edev - The device

    RETURNS:    TRUE if device is active; FALSE o/w

****************************************************************************/

u8 esas_device_is_active (PESASDEVICE edev)
{
    return ((edev != NULL)  && (edev->flags & EDF_IS_ACTIVE));
}


/****************************************************************************

    FUNCTION:   esas_device_set_active (PESASDEVICE edev)

    PURPOSE:    Set a device as active

    ARGUMENTS:  edev - The device

    RETURNS:    none

****************************************************************************/

void esas_device_set_active (PESASDEVICE edev)
{
    edev->flags |= EDF_IS_ACTIVE;
}


/****************************************************************************

    FUNCTION:   esas_device_set_inactive (PESASDEVICE edev)

    PURPOSE:    Set a device as inactive

    ARGUMENTS:  edev - The device

    RETURNS:    none

****************************************************************************/

void esas_device_set_inactive (PESASDEVICE edev)
{
    edev->flags &= ~EDF_IS_ACTIVE;
}





/****************************************************************************

    FUNCTION:   esas_port_create ()

    PURPOSE:    Create an esas port object

    ARGUMENTS:  NONE

    RETURNS:    The newly created esas port object.

****************************************************************************/

PESASPORT esas_port_create ()
{
    PESASPORT p;

    p = (PESASPORT) vmalloc (sizeof (ESASPORT));
    if (p == NULL)
    {
        printk (DEBUG_HEADER "Failed to allocate ESASPORT\n");
        return p;
    }

    memset (p, 0, sizeof (ESASPORT));

    return p;
}


/****************************************************************************

    FUNCTION:   esas_port_destroy (PESASPORT eport)

    PURPOSE:    Destroy an esas port object

    ARGUMENTS:  eport - the port object to destroy

    RETURNS:    NONE

****************************************************************************/

void esas_port_destroy (PESASPORT eport)
{
    if (eport)
        vfree(eport);
}


/****************************************************************************

    FUNCTION:   esas_port_add_host_phys (PESASTOPOLOGY etop, PESASPORT eport)

    PURPOSE:    Add host phys to the port.

    ARGUMENTS:  etop - The SAS topology
                eport - the port object to get the phys

    RETURNS:    The number of phys added; 0 means failure

****************************************************************************/

int esas_port_add_host_phys (PESASTOPOLOGY etop, PESASPORT eport)
{
    PESASDEVICE       ehost = eport->near_device;
    u8                phymask;
    u8                i;
    struct sas_phy    *phy;
    int               phycount = 0;
    ESAS_SMP_DISCOVER hostphy;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, ehost->dev,
                  "Add host phys to port\n");

    /* Get the host phy mask used to connect to the target */

    /* NOTE: For 3Gb, this will only return a single phy, even though */
    /* more phys may exist in the port. This is a known limitation.   */
    phymask = (u8)OsTransport (etop->host_no,
                               ATTO_TRANSPORT_HOST_GET_PHY_MASK,
                               eport->far_device->target_id, NULL);

    for (i = 0; i < ehost->num_phys; ++i)
    {
        if ((1 << i) & phymask)
        {
            phy = ehost->phy_list[i];

            if (phy == NULL)
            {
                /* Previous effort to allocate/configure this PHY failed. */
                printk (DEBUG_HEADER "PHY Discover failed for NULL PHY 0x%02x\n",i);
                continue;
            }

            if (OsTransport (etop->host_no,
                             ATTO_TRANSPORT_HOST_PHY_DISCOVER,
                             i,
                             &hostphy))
            {
                printk (DEBUG_HEADER "PHY Discover failed for PHY 0x%02x\n",i);
                continue;
            }

            if (atto_linkrate_to_sas_linkrate (hostphy.NegotiatedLogicalLinkRate)
            ==  SAS_LINK_RATE_UNKNOWN)
            {
                ATTO_DEV_LOG (KERN_WARN, ATTO_LOG_TOPOL, ehost->dev,
                              "PHY link rate 0x%02x unknown for phy 0x%02x\n",
                              hostphy.NegotiatedLogicalLinkRate, i);

                continue;
            }

            phy->negotiated_linkrate = atto_linkrate_to_sas_linkrate (hostphy.NegotiatedLogicalLinkRate);

            ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, ehost->dev,
                          "sas_port_add_phys(%d) called.\n", i);
            sas_port_add_phy (eport->port, phy);

            ++phycount;
        }
    }

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, ehost->dev,
                  "Added %d host phys to port\n", phycount);

    return phycount;
}


/****************************************************************************

    FUNCTION:   esas_port_add_expander_phys (PESASTOPOLOGY etop,
                                             PESASPORT eport)

    PURPOSE:    Add expander phys to the port.

    ARGUMENTS:  etop - The SAS topology
                eport - the port object to get the phys

    RETURNS:    The number of phys added; 0 means failure

****************************************************************************/

int esas_port_add_expander_phys (PESASTOPOLOGY etop, PESASPORT eport)
{
    int               phycount = 0;
    int               i;
    PESASDEVICE       exp = eport->near_device;
    ESAS_SMP_DISCOVER sd;
    struct sas_phy    *phy;
    enum sas_linkrate neg_rate;

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, exp->dev,
                  "Add expander phys to port 0x%llx<-->0x%llx\n",
                  eport->near_device->sas_address,
                  eport->far_device->sas_address);

    for (i = 0; i < exp->num_phys; ++i)
    {
        memset (&sd, 0, sizeof (ESAS_SMP_DISCOVER));
        sd.PhyId = i;
        sd.SasAddress = exp->sas_address;

        if (OsTransport (etop->host_no,
                         ATTO_TRANSPORT_SMP_DISCOVER,
                         exp->target_id,
                         &sd))
        {
            printk (DEBUG_HEADER "SMP Discover failed for PHY 0x%x on device 0x%llx.\n", i, exp->sas_address);
            continue;
        }

        if (atto_linkrate_to_sas_linkrate (sd.NegotiatedLogicalLinkRate)
        ==  SAS_LINK_RATE_UNKNOWN)
        {
            ATTO_DEV_LOG (KERN_WARN, ATTO_LOG_TOPOL, exp->dev,
                          "PHY link rate 0x%02x unknown for phy 0x%02x\n",
                          sd.NegotiatedLogicalLinkRate, i);

            continue;
        }

        if (sd.AttachedSasAddress == eport->far_device->sas_address)
        {
            ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, exp->dev,
                          "adding exp phy %d to port\n", i);
            phy = exp->phy_list[i];

            if (phy == NULL)
            {
                /* Previous effort to allocate/configure this PHY failed. */
                printk (DEBUG_HEADER "PHY SMP Discover failed for NULL PHY 0x%02x\n",i);
                continue;
            }

            neg_rate = atto_linkrate_to_sas_linkrate (sd.NegotiatedLogicalLinkRate);
            phy->negotiated_linkrate = neg_rate;

            sas_port_add_phy (eport->port, phy);

            ++phycount;
        }
    }

    ATTO_DEV_LOG (KERN_INFO, ATTO_LOG_TOPOL, exp->dev,
                  "Added %d expander phys\n", phycount);

    return phycount;
}


/****************************************************************************

    FUNCTION:   esas_port_get_near_phy_info (PESASTOPOLOGY etop,
                                             PESASPORT eport,
                                             PESAS_SMP_DISCOVER psd)

    PURPOSE:    Get info on the phy from the near side of the port

    ARGUMENTS:  etop - The SAS topology
                eport - the port object to get the phys
                psd - Pointer to SMP discover object that stores phy info

    RETURNS:    PHY index, if used; -1 on error

****************************************************************************/

int esas_port_get_near_phy_info (PESASTOPOLOGY etop,
                                 PESASPORT eport,
                                 PESAS_SMP_DISCOVER psd)
{
    int i;
    u8 phymask;
    int result = -1;

    /* If near device is host, then get PHY info for a host phy */
    /* that is attached to far device.                          */
    if (esas_device_is_host(eport->near_device))
    {
        /* Get the host phy mask used to connect to the far device */
        phymask = (u8)OsTransport (etop->host_no,
                                   ATTO_TRANSPORT_HOST_GET_PHY_MASK,
                                   eport->far_device->target_id,
                                   NULL);

        /* Get the PHY info for the first PHY in the mask */
        for (i = 0; i < eport->near_device->num_phys; ++i)
        {
            memset (psd, 0, sizeof(ESAS_SMP_DISCOVER));

            if ((1 << i) & phymask)
            {
                if (OsTransport (etop->host_no,
                                 ATTO_TRANSPORT_HOST_PHY_DISCOVER,
                                 i,
                                 psd))
                {
                    printk (DEBUG_HEADER "PHY Discover failed for PHY 0x%02x\n", i);
                    continue;

                }
                else
                {
                    result = i;
                    break;
                }
            }
        }
    }
    else
    {
        /* If near device is not host, find a phy on it that is attached    */
        /* to the far device.                                               */

        for (i = 0; i < eport->near_device->num_phys; ++i)
        {
            memset (psd, 0, sizeof (ESAS_SMP_DISCOVER));
            psd->PhyId = i;
            psd->SasAddress = eport->near_device->sas_address;

            if (OsTransport(etop->host_no,
                            ATTO_TRANSPORT_SMP_DISCOVER,
                            eport->near_device->target_id,
                            psd))
            {
                printk (DEBUG_HEADER "SMP Discover failed for PHY 0x%02x on device 0x%llx\n", i, eport->near_device->sas_address);
                continue;
            }

            if (psd->AttachedSasAddress == eport->far_device->sas_address)
            {
                result = i;
                break;
            }
        }
    }

    return result;
}


/****************************************************************************

    FUNCTION:   esas_port_by_sas_address (PESASDEVICE edev
                                          u64 sas_address)

    PURPOSE:    Find a port in the devices port list by sas address

    ARGUMENTS:  edev - The device who's port list to search
                sas_address - the sas addess to match in the far device

    RETURNS:    Pointer to port, if found; otherwise NULL

****************************************************************************/

PESASPORT esas_port_by_sas_address (PESASDEVICE edev,
                                    u64 sas_address)
{
    PESASPORT p;

    list_for_each_entry (p, &edev->port_list, list)
        if ((p->far_device) && (p->far_device->sas_address == sas_address))
            return p;

    return NULL;
}


/****************************************************************************

    FUNCTION:   atto_proto_to_sas_proto (u32 atto_proto, u8 is_target)

    PURPOSE:    Map ATTO protocol to SAS protocol

    ARGUMENTS:  atto_proto - The ATTO protocols (ATTO_PROTO_XXX)
                is_target - flag indicating whether atto_proto is
                            target protocol (1) or initiator protocol (0)

    RETURNS:    The SAS protocol

****************************************************************************/

enum sas_protocol atto_proto_to_sas_proto (u32 atto_proto,
                                           u8 is_target)
{
    enum sas_protocol p = 0;

    if (is_target)
    {
        if (atto_proto & ATTO_PROTO_TARG_SSP)
            p |= SAS_PROTOCOL_SSP;
        if (atto_proto & ATTO_PROTO_TARG_STP)
            p |= SAS_PROTOCOL_STP;
        if (atto_proto & ATTO_PROTO_TARG_SMP)
            p |= SAS_PROTOCOL_SMP;
        if (atto_proto & ATTO_PROTO_TARG_SATA)
            p |= SAS_PROTOCOL_SATA;
    }
    else
    {
        if (atto_proto & ATTO_PROTO_INIT_SSP)
            p |= SAS_PROTOCOL_SSP;
        if (atto_proto & ATTO_PROTO_INIT_STP)
            p |= SAS_PROTOCOL_STP;
        if (atto_proto & ATTO_PROTO_INIT_SMP)
            p |= SAS_PROTOCOL_SMP;
        if (atto_proto & ATTO_PROTO_INIT_SATA)
            p |= SAS_PROTOCOL_SATA;
    }

    return p;
}


/****************************************************************************

    FUNCTION:   atto_linkrate_to_sas_linkrate (u8 atto_rate)

    PURPOSE:    Map ATTO linkrate to SAS linkrate

    ARGUMENTS:  atto_rate - The ATTO link rate (ATTO_LINKRATE_XXX)

    RETURNS:    The SAS link rate

****************************************************************************/

enum sas_linkrate atto_linkrate_to_sas_linkrate (u8 atto_rate)
{
    enum sas_linkrate r;

    switch (atto_rate)
    {
    case ATTO_LINKRATE_1_5GB:
        r = SAS_LINK_RATE_1_5_GBPS;
        break;
    case ATTO_LINKRATE_3GB:
        r = SAS_LINK_RATE_3_0_GBPS;
        break;
    case ATTO_LINKRATE_6GB:
        r = SAS_LINK_RATE_6_0_GBPS;
        break;
    default:
        r = SAS_LINK_RATE_UNKNOWN;
        break;
    }

    return r;
}




#ifdef __VMKLNX__
/* See the comment in the VMware DDK scsi_transport_sas.h regarding  */
/* the need for this entry point. There's other code in the VMkernel */
/* that currently requires this method to be defined and working.    */
static int esashba_get_target_sas_identifier (struct scsi_target *starget, u64 *sas_id)
{
    struct sas_rphy *rphy = (struct sas_rphy *)starget->dev.parent;
    if (!rphy)
        return FAILED;
    *sas_id = rphy->identify.sas_address;
    return SUCCESS;
}

#endif /* __VMKLNX__ */

#endif /* ATTO_SAS_PORTS */

#endif /* ATTO_TRANSPORT */
